@@ -20,13 +20,16 @@
 */
 
 #include <config.h>
+#include "gtkhtml-compat.h"
+
 #include <gnome.h>
 #include <ctype.h>
 #include <libgnomeprint/gnome-print.h>
-#include <unicode.h>
+#include <gal/unicode/gunicode.h>
 
 #include "htmlembedded.h"
 #include "gtkhtml-embedded.h"
+#include "htmlfontmanager.h"
 #include "htmlprinter.h"
 
 /* #define PRINTER_DEBUG */
@@ -41,8 +44,8 @@ static HTMLPainterClass *parent_class = NULL;
 /* The following macros are used to convert between the HTMLEngine coordinate
    system (which uses integers) to the GnomePrint one (which uses doubles). */
 
-#define SCALE_ENGINE_TO_GNOME_PRINT(x) ((x) / 1024.0)
-#define SCALE_GNOME_PRINT_TO_ENGINE(x) ((gint) ((x) * 1024.0 + 0.5))
+#define SCALE_ENGINE_TO_GNOME_PRINT(x) ((printer->scale * (x)) / 1024.0)
+#define SCALE_GNOME_PRINT_TO_ENGINE(x) ((gint) (((x) * 1024.0 / printer->scale) + 0.5))
 
 /* Hm, this might need fixing.  */
 #define SPACING_FACTOR 1.2
@@ -64,14 +67,14 @@ insure_paper (void)
 }
 
 static gdouble
-get_page_height (HTMLPrinter *printer)
+printer_get_page_height (HTMLPrinter *printer)
 {
 	insure_paper ();
 	return gnome_paper_psheight (paper);
 }
 
 static gdouble
-get_page_width (HTMLPrinter *printer)
+printer_get_page_width (HTMLPrinter *printer)
 {
 	insure_paper ();
 	return gnome_paper_pswidth (paper);
@@ -110,7 +113,7 @@ get_bmargin (HTMLPrinter *printer)
 
 
 gdouble
-html_printer_scale_to_gnome_print (gint x)
+html_printer_scale_to_gnome_print (HTMLPrinter *printer, gint x)
 {
 	return SCALE_ENGINE_TO_GNOME_PRINT (x);
 }
@@ -126,7 +129,7 @@ html_printer_coordinates_to_gnome_print (HTMLPrinter *printer,
 	print_y = SCALE_ENGINE_TO_GNOME_PRINT (engine_y);
 
 	print_x = print_x + get_lmargin (printer);
-	print_y = (get_page_height (printer) - print_y) - get_tmargin (printer);
+	print_y = (printer_get_page_height (printer) - print_y) - get_tmargin (printer);
 
 	*print_x_return = print_x;
 	*print_y_return = print_y;
@@ -182,6 +185,7 @@ begin (HTMLPainter *painter,
 	pc      = printer->print_context;
 	g_return_if_fail (pc);
 
+	gnome_print_beginpage (pc, "page");
 	gnome_print_gsave (pc);
 
 	html_printer_coordinates_to_gnome_print (printer, x1, y1, &printer_x1, &printer_y1);
@@ -195,6 +199,7 @@ begin (HTMLPainter *painter,
 	gnome_print_lineto (pc, printer_x2, printer_y2);
 	gnome_print_lineto (pc, printer_x2, printer_y1);
 	gnome_print_lineto (pc, printer_x1, printer_y1);
+	gnome_print_closepath (pc);
 
 #ifdef PRINTER_DEBUG
 	gnome_print_gsave (pc);
@@ -262,7 +267,8 @@ get_black (const HTMLPainter *painter)
 static void
 prepare_rectangle (HTMLPainter *painter, gint _x, gint _y, gint w, gint h)
 {
-	GnomePrintContext *context = HTML_PRINTER (painter)->print_context;
+	HTMLPrinter *printer = HTML_PRINTER (painter);
+	GnomePrintContext *context = printer->print_context;
 	gdouble x;
 	gdouble y;
 	gdouble width;
@@ -279,12 +285,14 @@ prepare_rectangle (HTMLPainter *painter, gint _x, gint _y, gint w, gint h)
 	gnome_print_lineto  (context, x + width, y - height);
 	gnome_print_lineto  (context, x, y - height);
 	gnome_print_lineto  (context, x, y);
+	gnome_print_closepath (context);
 }
 
 static void
 do_rectangle (HTMLPainter *painter, gint x, gint y, gint w, gint h, gint lw)
 {
-	GnomePrintContext *context = HTML_PRINTER (painter)->print_context;
+	HTMLPrinter *printer = HTML_PRINTER (painter);
+	GnomePrintContext *context = printer->print_context;
 
 	gnome_print_setlinewidth (context, SCALE_ENGINE_TO_GNOME_PRINT (lw) * PIXEL_SIZE);
 	prepare_rectangle (painter, x, y, w, h);
@@ -342,7 +350,8 @@ draw_panel (HTMLPainter *painter,
 	    GtkHTMLEtchStyle inset,
 	    gint bordersize)
 {
-	GnomePrintContext *pc = HTML_PRINTER (painter)->print_context;
+	HTMLPrinter *printer = HTML_PRINTER (painter);
+	GnomePrintContext *pc = printer->print_context;
 	GdkColor *col1 = NULL, *col2 = NULL;
 	GdkColor dark, light;
 	gdouble x;
@@ -394,6 +403,7 @@ draw_panel (HTMLPainter *painter,
 	gnome_print_lineto  (pc, x + bs, y - bs );
 	gnome_print_lineto  (pc, x + bs, y - height + bs);
 	gnome_print_lineto  (pc, x, y - height);
+	gnome_print_closepath (pc);
 	gnome_print_fill    (pc);
 
 	if (col1)
@@ -406,6 +416,7 @@ draw_panel (HTMLPainter *painter,
 	gnome_print_lineto  (pc, x + width - bs, y - bs);
 	gnome_print_lineto  (pc, x + width - bs, y - height + bs);
 	gnome_print_lineto  (pc, x + bs, y - height + bs);
+	gnome_print_closepath (pc);
 	gnome_print_fill    (pc);
 }
 
@@ -439,6 +450,7 @@ draw_background (HTMLPainter *painter,
 		gnome_print_lineto (pc, x + width, y - height);
 		gnome_print_lineto (pc, x, y - height);
 		gnome_print_lineto (pc, x, y);
+		gnome_print_closepath (pc);
 
 		gnome_print_fill (pc);
 	}
@@ -497,6 +509,7 @@ fill_rect (HTMLPainter *painter,
 	gnome_print_lineto (printer->print_context, printer_x + printer_width, printer_y - printer_height);
 	gnome_print_lineto (printer->print_context, printer_x, printer_y - printer_height);
 	gnome_print_lineto (printer->print_context, printer_x, printer_y);
+	gnome_print_closepath (printer->print_context);
 
 	gnome_print_fill (printer->print_context);
 }
@@ -510,8 +523,7 @@ draw_text (HTMLPainter *painter,
 	GnomeFont *font;
 	HTMLPrinter *printer;
 	gdouble print_x, print_y;
-	gchar *text_tmp;
-	gint bytes;
+	gchar *end, *text_tmp;
 
 	printer = HTML_PRINTER (painter);
 	g_return_if_fail (printer->print_context != NULL);
@@ -523,10 +535,10 @@ draw_text (HTMLPainter *painter,
 
 	/* Oh boy, this sucks so much.  The GnomePrint API could be improved to
            avoid this.  */
-	bytes = unicode_offset_to_index (text, len);
-	text_tmp = alloca (bytes + 1);
-	memcpy (text_tmp, text, bytes);
-	text_tmp [bytes] = '\0';
+	end = g_utf8_offset_to_pointer (text, len);
+	text_tmp = alloca (end - text + 1);
+	memcpy (text_tmp, text, end - text);
+	text_tmp [end - text] = '\0';
 
 	font = html_painter_get_font (painter, painter->font_face, painter->font_style);
 	gnome_print_setfont (printer->print_context, font);
@@ -543,7 +555,7 @@ draw_text (HTMLPainter *painter,
 		gnome_print_setlinewidth (printer->print_context, 1.0);
 		gnome_print_setlinecap (printer->print_context, GDK_CAP_BUTT);
 
-		text_width = gnome_font_get_width_string_n (font, text, len);
+		text_width = gnome_font_get_width_utf8_sized (font, text, len);
 		if (painter->font_style & GTK_HTML_FONT_STYLE_UNDERLINE) {
 			descender = gnome_font_get_descender (font);
 			y = print_y + gnome_font_get_underline_position (font);
@@ -662,7 +674,7 @@ calc_text_width (HTMLPainter *painter,
 	font = html_painter_get_font (painter, face, style);
 	g_return_val_if_fail (font != NULL, 0);
 
-	width = gnome_font_get_width_string_n (font, text, len);
+	width = gnome_font_get_width_utf8_sized (font, text, len);
 
 	return SCALE_GNOME_PRINT_TO_ENGINE (width);
 }
@@ -670,22 +682,35 @@ calc_text_width (HTMLPainter *painter,
 static guint
 get_pixel_size (HTMLPainter *painter)
 {
+	HTMLPrinter *printer = HTML_PRINTER (painter);
+
 	return SCALE_GNOME_PRINT_TO_ENGINE (PIXEL_SIZE);
 }
 
-static HTMLFont *
-alloc_font (HTMLPainter *painter, gchar *face, gdouble size, GtkHTMLFontStyle style)
+static inline gdouble
+get_font_size (HTMLPrinter *printer, gboolean points, gdouble size)
 {
+	return printer->scale * (points ? size / 10 : size);
+}
+
+static HTMLFont *
+alloc_font (HTMLPainter *painter, gchar *face, gdouble size, gboolean points, GtkHTMLFontStyle style)
+{
+	HTMLPrinter *printer = HTML_PRINTER  (painter);
 	GnomeFontWeight weight;
 	GnomeFont *font;
 	gboolean italic;
-	guchar *s = face;
+	gchar *family = NULL;
 
 	weight = (style & GTK_HTML_FONT_STYLE_BOLD) ? GNOME_FONT_BOLD : GNOME_FONT_BOOK;
 	italic = (style & GTK_HTML_FONT_STYLE_ITALIC);
 
 	/* gnome-print is case sensitive - need to be fixed */
-	if (s && *s) {
+	if (face && *face) {
+		gchar *s;
+
+		s = family = html_font_manager_get_attr (face, 2);
+
 		/* capitalize */
 		*s = toupper (*s);
 		s++;
@@ -695,10 +720,22 @@ alloc_font (HTMLPainter *painter, gchar *face, gdouble size, GtkHTMLFontStyle st
 		}
 	}
 
-	if (!(font = gnome_font_new_closest ((face) ? face : "Courier", weight, italic, size)))
-		return NULL;
+	font = gnome_font_new_closest (family ? family : (style & GTK_HTML_FONT_STYLE_FIXED ? "Courier" : "Helvetica"),
+				       weight, italic, get_font_size (printer, points, size));
+	g_free (family);
 
-	return html_font_new (font, gnome_font_get_width_string_n (font, " ", 1));
+	if (font == NULL) {
+		GList *family_list;
+
+		family_list = gnome_font_family_list ();
+		if (family_list && family_list->data) {
+			font = gnome_font_new_closest (family_list->data,
+						       weight, italic, get_font_size (printer, points, size));
+			gnome_font_family_list_free (family_list);
+		}
+	}
+
+	return font ? html_font_new (font, gnome_font_get_width_utf8_sized (font, " ", 1)) : NULL;
 }
 
 static void
@@ -713,6 +750,17 @@ unref_font (HTMLPainter *painter, HTMLFont *font)
 	gtk_object_unref (GTK_OBJECT (font->data));
 }
 
+static guint
+get_page_width (HTMLPainter *painter, HTMLEngine *e)
+{
+	return html_printer_get_page_width (HTML_PRINTER (painter));
+}
+
+static guint
+get_page_height (HTMLPainter *painter, HTMLEngine *e)
+{
+	return html_printer_get_page_height (HTML_PRINTER (painter));
+}
 
 static void
 init (GtkObject *object)
@@ -721,6 +769,7 @@ init (GtkObject *object)
 
 	printer                = HTML_PRINTER (object);
 	printer->print_context = NULL;
+	printer->scale         = 1.0;
 }
 
 static void
@@ -756,6 +805,8 @@ class_init (GtkObjectClass *object_class)
 	painter_class->get_pixel_size = get_pixel_size;
 	painter_class->set_clip_rectangle = set_clip_rectangle;
 	painter_class->draw_embedded = draw_embedded;
+	painter_class->get_page_width = get_page_width;
+	painter_class->get_page_height = get_page_height;
 
 	parent_class = gtk_type_class (html_painter_get_type ());
 }
@@ -807,7 +858,7 @@ html_printer_get_page_width (HTMLPrinter *printer)
 	g_return_val_if_fail (printer != NULL, 0);
 	g_return_val_if_fail (HTML_IS_PRINTER (printer), 0);
 
-	printer_width = get_page_width (printer) - get_lmargin (printer) - get_rmargin (printer);
+	printer_width = printer_get_page_width (printer) - get_lmargin (printer) - get_rmargin (printer);
 	engine_width = SCALE_GNOME_PRINT_TO_ENGINE (printer_width);
 
 	return engine_width;
@@ -822,7 +873,7 @@ html_printer_get_page_height (HTMLPrinter *printer)
 	g_return_val_if_fail (printer != NULL, 0);
 	g_return_val_if_fail (HTML_IS_PRINTER (printer), 0);
 
-	printer_height = get_page_height (printer) - get_lmargin (printer) - get_rmargin (printer);
+	printer_height = printer_get_page_height (printer) - get_lmargin (printer) - get_rmargin (printer);
 	engine_height = SCALE_GNOME_PRINT_TO_ENGINE (printer_height);
 
 	return engine_height;
