@@ -26,9 +26,11 @@
 #include <libart_lgpl/art_rect.h>
 #include <string.h>
 #include "htmlcluev.h"
+#include "htmlengine-edit.h"
 #include "htmlengine-save.h"
 #include "htmlpainter.h"
 #include "htmltable.h"
+#include "htmltablepriv.h"
 #include "htmltablecell.h"
 
 /* FIXME: This always behaves as a transparent object, even when it
@@ -46,10 +48,12 @@ draw_background_helper (HTMLTableCell *cell,
 			gint tx, gint ty)
 {
 	HTMLObject *o;
+	HTMLTable  *t;
 	GdkPixbuf  *pixbuf = NULL;
 	GdkColor   *color = NULL;
 
 	o = HTML_OBJECT (cell);
+	t = HTML_IS_TABLE (o->parent) ? HTML_TABLE (o->parent) : NULL;
 
 	if (cell->have_bg) {
 		if (! cell->bg_allocated) {
@@ -57,13 +61,17 @@ draw_background_helper (HTMLTableCell *cell,
 			cell->bg_allocated = TRUE;
 		}
 		color = &cell->bg;
+	} else if (t && t->bgColor) {
+		html_painter_alloc_color (p, t->bgColor);
+		color = t->bgColor;
 	}
 
 	if (cell->have_bgPixmap) {
 		if (cell->bgPixmap->pixbuf) {
 			pixbuf = cell->bgPixmap->pixbuf;
 		}
-	}
+	} else if (t && t->bgPixmap)
+		pixbuf = t->bgPixmap->pixbuf;
 
 	html_painter_draw_background (p,
 				      color,
@@ -100,35 +108,41 @@ copy (HTMLObject *self, HTMLObject *dest)
 }
 
 static gboolean
-merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList *left, GList *right)
+merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList **left, GList **right, HTMLCursor *cursor)
 {
 	HTMLTableCell *c1 = HTML_TABLE_CELL (self);
 	HTMLTableCell *c2 = HTML_TABLE_CELL (with);
 
-	printf ("merge cells %d,%d %d,%d\n", c1->row, c1->col, c2->row, c2->col);
+	/* g_print ("merge cells %d,%d %d,%d\n", c1->row, c1->col, c2->row, c2->col); */
 
-	return HTML_OBJECT_TYPE (with) == HTML_TYPE_CLUEV || c1->col == c2->col
-		? (* HTML_OBJECT_CLASS (parent_class)->merge) (self, with, e, left, right)
-		: FALSE;
+	if (HTML_OBJECT_TYPE (with) == HTML_TYPE_CLUEV || (c1->col == c2->col && c1->row == c2->row)) {
+		gboolean rv;
+		rv = (* HTML_OBJECT_CLASS (parent_class)->merge) (self, with, e, left, right, cursor);
+		if (rv && with->parent && HTML_IS_TABLE (with->parent)) {
+			self->next = NULL;
+			html_object_remove_child (with->parent, with);
+			/* FIXME spanning */
+			html_table_set_cell (HTML_TABLE (self->parent),
+					     HTML_TABLE_CELL (self)->row, HTML_TABLE_CELL (self)->col,
+					     HTML_TABLE_CELL (self));
+		}
+
+		return rv;
+	} else
+		return FALSE;
 }
 
 static gint
 calc_min_width (HTMLObject *o,
 		HTMLPainter *painter)
 {
-	HTMLObject *obj;
-	gint minWidth;
+	if (HTML_TABLE_CELL (o)->no_wrap)
+		return MAX ((* HTML_OBJECT_CLASS (parent_class)->calc_preferred_width) (o, painter),
+			    o->flags & HTML_OBJECT_FLAG_FIXEDWIDTH
+			    ? HTML_TABLE_CELL (o)->fixed_width * html_painter_get_pixel_size (painter)
+			    : 0);
 
-	minWidth = 0; 
-	for (obj = HTML_CLUE (o)->head; obj != NULL; obj = obj->next) {
-		gint w;
-
-		w = html_object_calc_min_width (obj, painter);
-		if (w > minWidth)
-			minWidth = w;
-	}
-
-	return minWidth;
+	return (* HTML_OBJECT_CLASS (parent_class)->calc_min_width) (o, painter);
 }
 
 static gint
@@ -136,8 +150,8 @@ calc_preferred_width (HTMLObject *o,
 		      HTMLPainter *painter)
 {
 	return o->flags & HTML_OBJECT_FLAG_FIXEDWIDTH
-		? MAX (html_object_calc_min_width (o, painter), HTML_TABLE_CELL (o)->fixed_width
-		       * html_painter_get_pixel_size (painter))
+		? MAX (html_object_calc_min_width (o, painter), 
+		       HTML_TABLE_CELL (o)->fixed_width * html_painter_get_pixel_size (painter))
 		: (* HTML_OBJECT_CLASS (parent_class)->calc_preferred_width) (o, painter);
 }
 
@@ -173,13 +187,17 @@ clue_move_children (HTMLClue *clue, gint x_delta, gint y_delta)
 }
 
 static gboolean
-calc_size (HTMLObject *o, HTMLPainter *painter)
+calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs)
 {
 	HTMLTableCell *cell;
 	gboolean rv;
+	gint old_width, old_height;
+
+	old_width  = o->width;
+	old_height = o->ascent + o->descent;
 
 	cell = HTML_TABLE_CELL (o);
-	rv   = (* HTML_OBJECT_CLASS (parent_class)->calc_size) (o, painter);
+	rv   = (* HTML_OBJECT_CLASS (parent_class)->calc_size) (o, painter, changed_objs);
 
 	if (cell->fixed_height && o->ascent + o->descent < cell->fixed_height) {
 		gint remains = cell->fixed_height - (o->ascent + o->descent);
@@ -189,7 +207,7 @@ calc_size (HTMLObject *o, HTMLPainter *painter)
 		switch (HTML_CLUE (o)->valign) {
 		case HTML_VALIGN_TOP:
 			break;
-		case HTML_VALIGN_CENTER:
+		case HTML_VALIGN_MIDDLE:
 			clue_move_children (HTML_CLUE (o), 0, remains >> 1);
 			break;
 		case HTML_VALIGN_NONE:
@@ -202,6 +220,9 @@ calc_size (HTMLObject *o, HTMLPainter *painter)
 		rv = TRUE;
 	}
 
+	if (o->parent && (o->width != old_width || o->ascent + o->descent != old_height))
+		html_object_add_to_changed (changed_objs, o->parent);
+
 	return rv;
 }
 
@@ -262,20 +283,50 @@ save (HTMLObject *self,
 {
 	HTMLTableCell *cell = HTML_TABLE_CELL (self);
 
-	SB "<TD" SE;
+	SB cell->heading ? "<TH" : "<TD" SE;
 	if (cell->have_bg
 	    && (!HTML_TABLE (self->parent)->bgColor || !gdk_color_equal (&cell->bg, HTML_TABLE (self->parent)->bgColor)))
 		SB " BGCOLOR=\"#%02x%02x%02x\"",
 			cell->bg.red >> 8,
 			cell->bg.green >> 8,
 			cell->bg.blue >> 8 SE;
-	if (self->percent > 0) {
-		SB " WIDTH=\"%d%%\"", self->percent SE;
+	
+	if (cell->have_bgPixmap) {
+		gchar * url = html_image_resolve_image_url (state->engine->widget, cell->bgPixmap->url);
+		SB " BACKGROUND=\"%s\"", url SE;
+		g_free (url);
+	}
+	if (cell->cspan != 1)
+		SB " COLSPAN=\"%d\"", cell->cspan SE;
+
+	if (cell->rspan != 1)
+		SB " ROWSPAN=\"%d\"", cell->rspan SE;
+
+	if (cell->percent_width) {
+		SB " WIDTH=\"%d%%\"", cell->fixed_width SE;
 	} else if (self->flags & HTML_OBJECT_FLAG_FIXEDWIDTH)
 		SB " WIDTH=\"%d\"", cell->fixed_width SE;
+	if (cell->no_wrap)
+		SB " NOWRAP" SE;
+	if (HTML_CLUE (cell)->halign != HTML_HALIGN_NONE)
+		SB " ALIGN=\"%s\"",
+                   html_engine_save_get_paragraph_align (html_alignment_to_paragraph (HTML_CLUE (cell)->halign)) SE;
+	if (HTML_CLUE (cell)->valign != HTML_VALIGN_MIDDLE)
+		SB " VALIGN=\"%s\"",
+                   HTML_CLUE (cell)->valign == HTML_VALIGN_TOP ? "top" : "bottom" SE;
 	SB ">\n" SE;
 	if (!(*HTML_OBJECT_CLASS (parent_class)->save) (self, state))
 		return FALSE;
+
+	/*
+	  save at least &nbsp; for emty cell
+	  it's not really clean behavior, but I think it's better than saving empty cells.
+	  let me know if you have better idea
+
+	if (HTML_CLUE (self)->head && HTML_CLUE (self)->head == HTML_CLUE (self)->tail
+	    && HTML_IS_CLUEFLOW (HTML_CLUE (self)->head) && html_clueflow_is_empty (HTML_CLUEFLOW (HTML_CLUE (self)->head)))
+	    SB "&nbsp;" SE; */
+
 	SB "</TD>\n" SE;
 
 	return TRUE;
@@ -319,7 +370,6 @@ html_table_cell_class_init (HTMLTableCellClass *klass,
 void
 html_table_cell_init (HTMLTableCell *cell,
 		      HTMLTableCellClass *klass,
-		      gint percent,
 		      gint rs, gint cs,
 		      gint pad)
 {
@@ -331,17 +381,18 @@ html_table_cell_init (HTMLTableCell *cell,
 	cluev = HTML_CLUEV (cell);
 	clue = HTML_CLUE (cell);
 
-	html_cluev_init (cluev, HTML_CLUEV_CLASS (klass), 0, 0, percent);
+	html_cluev_init (cluev, HTML_CLUEV_CLASS (klass), 0, 0, 0);
 
 	object->flags &= ~HTML_OBJECT_FLAG_FIXEDWIDTH;
 
-	clue->valign = HTML_VALIGN_BOTTOM;
-	clue->halign = HTML_HALIGN_LEFT;
+	clue->valign = HTML_VALIGN_MIDDLE;
+	clue->halign = HTML_HALIGN_NONE;
 
 	cell->fixed_width  = 0;
 	cell->fixed_height = 0;
+	cell->percent_width = FALSE;
+	cell->percent_height = FALSE;
 
-	cell->padding = 0;
 	cluev->padding = pad;
 	cell->rspan = rs;
 	cell->cspan = cs;
@@ -351,32 +402,38 @@ html_table_cell_init (HTMLTableCell *cell,
 	cell->have_bg = FALSE;
 	cell->have_bgPixmap = FALSE;
 	cell->bg_allocated = FALSE;
+
+	cell->no_wrap = FALSE;
+	cell->heading = FALSE;
 }
 
 HTMLObject *
-html_table_cell_new (gint percent,
-		     gint rs, gint cs,
-		     gint pad)
+html_table_cell_new (gint rs, gint cs, gint pad)
 {
 	HTMLTableCell *cell;
 
 	cell = g_new (HTMLTableCell, 1);
-	html_table_cell_init (cell, &html_table_cell_class, percent, rs, cs, pad);
+	html_table_cell_init (cell, &html_table_cell_class, rs, cs, pad);
 
 	return HTML_OBJECT (cell);
 }
 
 void
-html_table_cell_set_fixed_width (HTMLTableCell *cell, gint width)
+html_table_cell_set_fixed_width (HTMLTableCell *cell, gint width, gboolean percented)
 {
-	HTML_OBJECT (cell)->flags |= HTML_OBJECT_FLAG_FIXEDWIDTH;
+	if (percented)
+		HTML_OBJECT (cell)->flags &= ~HTML_OBJECT_FLAG_FIXEDWIDTH;
+	else
+		HTML_OBJECT (cell)->flags |= HTML_OBJECT_FLAG_FIXEDWIDTH;
 	cell->fixed_width = width;
+	cell->percent_width = percented;
 }
 
 void
-html_table_cell_set_fixed_height (HTMLTableCell *cell, gint height)
+html_table_cell_set_fixed_height (HTMLTableCell *cell, gint height, gboolean percented)
 {
 	cell->fixed_height = height;
+	cell->percent_height = percented;
 }
 
 void
